@@ -1,6 +1,8 @@
 from heppy.value import Value
 from heppy.histogram import *
-from heppy.readroot import *
+from heppy.readroot import * # uses uproot
+import heppy.readroot_legacy # uses ROOT and root-numpy
+from heppy.readtrex import readtrex
 from heppy.panel import *
 from heppy.plot import *
 from heppy.heatmap import *
