@@ -1,4 +1,5 @@
 import numpy as np
+import numbers
 import operator
 import warnings
 from copy import deepcopy
@@ -186,9 +187,9 @@ class basehistogram(object):
 
     def set_heights(self, heights):
         '''
-        Set bin heights
+        Set bin heights to an array of the same dimension as the current areas or to a scalar
         '''
-        assert heights.shape == self.binsizes.shape
+        assert heights.shape == self.binsizes.shape or isinstance(heights, numbers.Number)
         self.areas = self.binsizes * heights
 
 
@@ -217,11 +218,11 @@ class basehistogram(object):
 
     def net_variations(self, variations='all', subtract_nominal=False, relative=False):
         '''
-        Return upper and lower net heights variation of the histogram as a tuple.
+        Return upper and lower net areas variation of the histogram as a tuple.
 
         @variations should be a sequence of considered variation names or the string 'all'
-        @subtract_nominal: if True, return the differences with respect to the nominal heights
-        @relative: if True, divide by the nominal heights
+        @subtract_nominal: if True, return the differences with respect to the nominal areas
+        @relative: if True, divide by the nominal areas
 
         CAUTION: this method cannot yet deal with systematic uncertainties for which the up- and down-shift lie on the same side of the nominal.
         This is because the variations are fundamentally treated independently of each other, so there is no sense of the up- and down-shift
@@ -516,6 +517,34 @@ class basehistogram(object):
 
 
 
+    def clip(self, minimum=None, maximum=None):
+        """Clips the bin areas at a minimum and/or maximum value.
+
+        Similar to ``numpy.clip`` (see `here <https://numpy.org/doc/stable/reference/generated/numpy.clip.html>`_).
+
+        Both the nominals and all correlated and uncorrelated variations of the histogram are clipped.
+        *Note that the net variation of multiple individual variations considered together may still lie
+        outside of the clip range!*
+
+        :param minimum: minimum accepted value; areas below this value will be set to the value.
+        If ``None``, no minimum is imposed.
+        :type minimum: ``float`` or ``None``
+        :param maximum: maximum accepted value; areas above this value will be set to the value.
+        If ``None``, no maximum is imposed.
+        :type maximum: ``float`` or ``None``
+
+        The behaviour is like that of ``numpy.clip``, e.g. with regards to what happens if ``minimum`` is
+        greater than ``maximum``.
+
+        **"Undocumented" feature:** you can actually pass a Numpy ``array_like`` as ``minimum`` or ``maximum``,
+        not just a single ``float``!
+        """
+        def clip_function(array):
+            return np.clip(array, minimum, maximum)
+        self.apply_inplace(clip_function)
+
+
+
     def __repr__(self):
         '''
         String representation of the histogram that can be used to recreate it, e.g. with the following code:
@@ -718,9 +747,13 @@ class histogram1d(basehistogram):
     def squash_highest_bin(self, squash_above, new_xmax):
         '''
         Merge all bins from @squash_above upwards and set the highest bin edge to @new_xmax.
+
+        Does nothing if squash_above is >= the highest bin edge.
         '''
         if squash_above >= new_xmax:
             raise RuntimeError('Setting new highest bin edge above the bin squashing threshold is forbidden')
+        if squash_above >= np.max(self.binedges):
+            return
         self.merge_bins(squash_above, np.max(self.binedges))
         # Only now update binedges:
         self.binedges[-1] = new_xmax
@@ -799,6 +832,36 @@ class histogram1d(basehistogram):
 
 
 
+    def cumulative(self, side='left', nan_to_num=True):
+        """Returns a new histogram containing the cumulative sums of the bin areas.
+
+        Analogous to Numpy's ``cumsum`` function.
+
+        Note: the plot attributes are copied over from ``self`` to the returned
+        histogram.
+
+
+        :param side: which side to integrate from, may be either ``'left'`` or ``'right'``
+        :type side: ``str``
+        :param nan_to_num: treat NaN as 0 when taking the cumulative sum?
+        :type nan_to_num: ``bool``
+
+        :returns: histogram of the cumulative sum of the bin areas
+        :rtype: :py:class:`heppy.histogram1d`
+        """
+        areas = deepcopy(self.areas)
+        if nan_to_num:
+            areas = np.nan_to_num(areas)
+        if side == 'left':
+            cumheights = np.cumsum(areas)
+        elif side == 'right':
+            cumheights = np.cumsum(areas[::-1])[::-1]
+        else:
+            raise RuntimeError('Argument "side" must be either "left" or "right".')
+        return histogram1d(self.binedges, cumheights, areas=False, plot_attributes=self.plot_attributes)
+
+
+
     def to_yoda(self, identifier, metadata={}):
         """Returns the histogram in YODA output format as a string.
 
@@ -817,6 +880,7 @@ class histogram1d(basehistogram):
             Add Variations and ErrorBreakdown fields
 
             Unit test this method
+
         """
         template = dedent('''\
         BEGIN YODA_SCATTER2D_V2 {identifier}
@@ -852,6 +916,96 @@ class histogram1d(basehistogram):
 
 
 
+    def to_root(self, nominal_label='nominal', key_form='{name}_{variation}', errors=None):
+        """Returns a dictionary of variation names mapped to ROOT TH1D's.
+
+        :param nominal_label: dictionary key to give to the nominal histogram
+        :type nominal_label: ``str``
+        :param key_form: template for keys in the returned dictionary. The
+            names of the histograms will also be set to these values. It accepts
+            the formatting fields `name` (for `self.name`) and variation (for
+            the name of the variation)
+        :type key_form: ``str``
+        :param errors: optional array of bin errors. Must have the same dimensionality
+        as the histogram areas. Note that the same errors are used for all variation
+        histograms.
+        :type errors: ``np.array`` or ``None``
+
+        :returns: dictionary of ROOT histograms (values) mapped by variation
+            name (keys)
+        :rtype: ``dict`` of ``str`` and ``ROOT.TH1D``
+        """
+        import ROOT
+        import root_numpy as rnp
+        from array import array
+        def to_th1(name, binedges, areas):
+            th1 = ROOT.TH1D(name, name, len(binedges)-1, array('d', [b for b in binedges]))
+            rnp.array2hist(areas, th1, errors=errors)
+            return th1
+        th1s = {}
+        # Add nominal
+        key = key_form.format(name=self.name, variation=nominal_label)
+        th1s[key] = to_th1(key, self.binedges, self.areas)
+        for variation, areas in {**self.uncorr_variations, **self.corr_variations}.items():
+            key = key_form.format(name=self.name, variation=variation)
+            th1s[key] = to_th1(key, self.binedges, areas)
+        return th1s
+
+
+
+    def to_root_file(self, filename, recreate=False, **kwargs):
+        """Writes this histogram to a ROOT file as a set of TH1D's.
+
+        :param filename: name or path to the ROOT file. May exist or not.
+        :type filename: ``str``
+        :param recreate: only has an effect if the ROOT file exists already. If
+            `True`, it will be overwritten. Otherwise, the new histograms will
+            be added to the existing file.
+        :type recreate: ``bool``
+        :param **kwargs: keyword arguments that get passed on to
+            :py:func:`histogram1d.to_root`
+        """
+        import ROOT
+        mode = 'RECREATE' if recreate else 'UPDATE'
+        rootfile = ROOT.TFile(filename, mode)
+        if not rootfile.IsOpen():
+            raise OSError(f'Cannot open ROOT file "{filepath}"')
+        th1s = self.to_root(**kwargs)
+        for key, th1 in th1s.items():
+            th1.Write(key)
+
+
+
+
+
+def zeros_like(a, name=None):
+    '''
+    Returns a histogram of zeros with the same bin edges and variations as ``a``.
+
+    The attributes and plot attributes are not kept.
+
+    :param a: the bin edges and variation names of ``a`` define these same
+    attributes of the returned histogram.
+    :type a: :py:class:`heppy.histogram1d` or :py:class:`heppy.histogram2d`
+    :param name: if given, this becomes the name attribute of the returned
+    histogram. If ``Non``, the returned histogram has the name ``"zeros like " + a.name``
+    :type name: ``str`` or ``None``
+
+    :returns: :py:class:`heppy.histogram1d` or :py:class:`heppy.histogram2d`
+    '''
+    b = deepcopy(a)
+    zeros = np.zeros_like(a.areas)
+    b.areas = np.zeros_like(a.areas)
+    for variation in b.uncorr_variations.keys():
+        b.uncorr_variations[variation] = np.zeros_like(a.areas)
+    for variation in b.corr_variations.keys():
+        b.corr_variations[variation] = np.zeros_like(a.areas)
+    b.name = f'zeros like {a.name}' if name is None else name
+    b.attributes = {}
+    b.plot_attributes = {}
+    return b
+
+
 
 
 
@@ -903,8 +1057,38 @@ def histdiv(a, b, corr=None, ignore_denominator_uncertainty=False):
     else:
         r = a / b
     # Undo division by bin sizes (areas=False):
-    r = histogram(r.binedges, r.areas, areas=False, name=r.name, uncorr_variations=r.uncorr_variations, corr_variations=r.corr_variations, attributes=r.attributes, plot_attributes=r.plot_attributes)
-    if corr:
+    r = type(a)(r.binedges, r.areas, areas=False, name=r.name, uncorr_variations=r.uncorr_variations, corr_variations=r.corr_variations, attributes=r.attributes, plot_attributes=r.plot_attributes)
+    if isinstance(corr, str) and corr.startswith('subset-toys'):
+        ntoys = int(corr.split('-')[-1])
+        if (ntoys < 100):
+            print('Warning: it is not recommended to use ntoys < 100! Uncertainty propagation will be inaccurate, including due to internal mechanism for dealing with bins in which numerator == denominator!')
+        both = a.areas
+        denonly = b.areas - a.areas
+        if (denonly < 0).any():
+            raise RuntimeError('Numerator is not a subset of denominator! Cannot use this uncertainty propagation method.')
+        both_toys = np.random.poisson(both, (ntoys, *both.shape))
+        denonly_toys = np.random.poisson(denonly, (ntoys, *both.shape))
+        ratio_toys = both_toys / (both_toys + denonly_toys)
+        #
+        # Compute final uncertainty as RMS distance from nominal (up and down separately)
+        #
+        shift_toys = ratio_toys - r.heights
+        upshift_toys = np.clip(shift_toys, 0.0, None)
+        downshift_toys = np.clip(shift_toys, None, 0.0)
+        # Need the effective number of toys separately for up and down
+        # To get it, we just count how many down and how many up fluctuations we had in each bin
+        n_effective_toys_up = np.sum(np.ceil(upshift_toys), axis=0)
+        n_effective_toys_down = np.sum(np.ceil(np.abs(downshift_toys)), axis=0)
+        # Note that in each bin, n_effective_toys_up and n_effective_toys_down sum to either ntoys or zero
+        # The latter happens if the numerator and denominator hists have exactly the same entries, and hence
+        # will fluctuate coherently and not incur any uncertainty
+        # To avoid dividing by zero in such cases, we set them to 1. It doesn't matter what we set them to,
+        # since the uncertainties will anyway be zero in these bins.
+        n_effective_toys_up = np.clip(n_effective_toys_up, 1.0, None)
+        n_effective_toys_down = np.clip(n_effective_toys_down, 1.0, None)
+        r.uncorr_variations['Statistical__1up'] = r.areas + np.sqrt(np.sum(upshift_toys**2, axis=0) / n_effective_toys_up) * r.binsizes
+        r.uncorr_variations['Statistical__1down'] = r.areas - np.sqrt(np.sum(downshift_toys**2, axis=0) / n_effective_toys_down) * r.binsizes
+    elif corr is not None:
         raise NotImplementedError('Division of histograms that are not statistically independent still needs to be implemented.')
     return r
 
@@ -1007,6 +1191,8 @@ class histogram2d(basehistogram):
         shifts = var_areas - self.areas
         nominal = self.areas.copy() # Nominal areas to be rebinned along with the variation shifts
         for index in [0, 1]:
+            # print(self.binedges[index])
+            # print(newedges[index])
             shifts = sign * np.sqrt(np.add.reduceat(shifts**2, np.searchsorted(self.binedges[index], newedges[index])[:-1], axis=index))
             nominal = np.add.reduceat(nominal, np.searchsorted(self.binedges[index], newedges[index])[:-1], axis=index)
         return nominal + shifts
@@ -1035,7 +1221,7 @@ class histogram2d(basehistogram):
         for index in [0, 1]:
             bin_index_function = [self.bin_index_x, self.bin_index_y][index]
             min_index = bin_index_function(np.min(newedges[index]))
-            max_index = None if np.max(newedges[index]) == np.max(self.binedges[index]) else bin_index_function(np.max(newedges[index]))
+            max_index = None if np.isclose(np.max(newedges[index]), np.max(self.binedges[index])) else bin_index_function(np.max(newedges[index]))
             if min_index > 0 or max_index is not None:
                 if index == 0:
                     self.uncorr_variations = {name : areas[min_index:max_index,:] for name, areas in self.uncorr_variations.items()}
@@ -1085,7 +1271,7 @@ class histogram2d(basehistogram):
         Project histogram to one axis by integrating over the other. Correlated and uncorrelated
         uncertainties are computed for the resulting one-dimensional histogram.
 
-        :param axis: which axis to project onto
+        :param axis: which axis to project onto, i.e. the *axis that is kept*
         :type axis: :code:`'x'` or :code:`'y'`
 
         :param name: name for the projection histogram
@@ -1119,7 +1305,7 @@ class histogram2d(basehistogram):
         :type axis: :code:`'x'` or :code:`'y'`
 
         :param bin_index: index of the bin on the *axis that is not kept*
-        :type bin_index: ``float``
+        :type bin_index: ``int``
 
         :param name: name for the slice histogram
         :type name: ``str``
@@ -1144,17 +1330,59 @@ class histogram2d(basehistogram):
             'y' : (np.array([bin_lower_edge, bin_upper_edge]), sliced.binedges[1]),
             }[axis]
         sliced.rebin(newedges)
+        width_of_slice_bin = bin_upper_edge - bin_lower_edge
+        sliced = sliced * (1.0 / width_of_slice_bin) # is this correct? It makes the test fail --- maybe the test is wrong? How do we want to normalise?
         if name == '':
             name = self.name + ' slice along {axis}-axis in {other_axis}-bin [{binmin}, {binmax})'.format(axis=axis, other_axis=other_axis, binmin=bin_lower_edge, binmax=bin_upper_edge)
         return sliced.as_1d(name=name)
 
 
 
+    def height(self, bin_index_x, bin_index_y):
+        """Returns the height of the given bin indices with uncertainties.
+
+        :param bin_index_x: bin index along x-axis
+        :param bin_index_x: ``int``
+        :param bin_index_y: bin index along y-axis
+        :param bin_index_y: ``int``
+
+        :returns: height of the indexed bin including its variations
+        :rtype: :py:class:`heppy.value`
+        """
+        uncorr_variations = {key : (values/self.binsizes)[bin_index_x,bin_index_y] for key, values in self.uncorr_variations.items()}
+        corr_variations = {key : (values/self.binsizes)[bin_index_x,bin_index_y] for key, values in self.corr_variations.items()}
+        return Value(self.heights[bin_index_x,bin_index_y], uncorr_variations=uncorr_variations, corr_variations=corr_variations)
+
+
+
+    def iterheights(self, faster='y'):
+        """Generates iterator over heights.
+
+        :param faster: controls the iteration order by specifying along which axis the bin index changes faster
+        :type faster: ``str``; ``'x'`` or ``'y'``
+
+        :returns: bin heights including their variations
+        :rtype: :py:class:`heppy.value`
+        """
+        if not faster in ('x', 'y'):
+            raise ValueError('Valid values for keyword argument "faster" are: "x", "y"')
+        nbins_x, nbins_y = self.nbins
+        if faster == 'y':
+            for bin_index_x in range(nbins_x):
+                for bin_index_y in range(nbins_y):
+                    yield self.height(bin_index_x, bin_index_y)
+        elif faster == 'x':
+            for bin_index_y in range(nbins_y):
+                for bin_index_x in range(nbins_x):
+                    yield self.height(bin_index_x, bin_index_y)
+
+
+
     def iterbins(self):
         """Generates iterator over bins, yielding bin edges and heights.
 
-        :returns: x-axis bin egdes, y-axis bin edges, and nominal bin height
-        :rtype: ``tuple`` of the following: ``tuple`` of two ``float``, ``tuple`` of two ``float``, and one ``float``
+        :returns: x-axis bin egdes, y-axis bin edges, and bin height with variations
+        :rtype: ``tuple`` of the following: ``tuple`` of two ``float``, ``tuple`` of two ``float``, and one :py:class:`heppy.value`
 
         Usage example:
 
@@ -1170,7 +1398,7 @@ class histogram2d(basehistogram):
             >>> x = np.array([-7., 0., 5., 50.]) # bin edges in x
             >>> y = np.array([-1., 0., 1.])      # bin edges in y
             >>> h = hp.histogram2d((x, y), heights)
-            >>> for binedges_x, binedges_y, height in h.iterbins(): print(binedges_x, binedges_y, height)
+            >>> for binedges_x, binedges_y, height in h.iterbins(): print(binedges_x, binedges_y, height.nominal)
             (-7.0, 0.0) (-1.0, 0.0) 1.0
             (-7.0, 0.0) (0.0, 1.0) 5.0
             (0.0, 5.0) (-1.0, 0.0) 2.0
@@ -1185,12 +1413,72 @@ class histogram2d(basehistogram):
         bin_upper_edges_x = np.repeat(binedges_x[1:], nbins_y)
         bin_lower_edges_y = np.tile(binedges_y[:-1], nbins_x)
         bin_upper_edges_y = np.tile(binedges_y[1:], nbins_x)
-        expanded_heights = np.ravel(self.heights)
-        for bin_lower_edge_x, bin_upper_edge_x, bin_lower_edge_y, bin_upper_edge_y, height in zip(bin_lower_edges_x, bin_upper_edges_x, bin_lower_edges_y, bin_upper_edges_y, expanded_heights):
+        for bin_lower_edge_x, bin_upper_edge_x, bin_lower_edge_y, bin_upper_edge_y, height in zip(bin_lower_edges_x, bin_upper_edges_x, bin_lower_edges_y, bin_upper_edges_y, self.iterheights(faster='y')):
             yield (bin_lower_edge_x, bin_upper_edge_x), (bin_lower_edge_y, bin_upper_edge_y), height
 
 
 
+    def to_root(self, nominal_label='nominal', key_form='{name}_{variation}', errors=None):
+        """Returns a dictionary of variation names mapped to ROOT TH2D's.
+
+        :param nominal_label: dictionary key to give to the nominal histogram
+        :type nominal_label: ``str``
+        :param key_form: template for keys in the returned dictionary. The
+            names of the histograms will also be set to these values. It accepts
+            the formatting fields `name` (for `self.name`) and variation (for
+            the name of the variation)
+        :type key_form: ``str``
+        :param errors: optional array of bin errors. Must have the same dimensionality
+        as the histogram areas. Note that the same errors are used for all variation
+        histograms.
+        :type errors: ``np.array`` or ``None``
+
+
+        :returns: dictionary of ROOT histograms (values) mapped by variation
+            name (keys)
+        :rtype: ``dict`` of ``str`` and ``ROOT.TH2D``
+        """
+        import ROOT
+        import root_numpy as rnp
+        from array import array
+        def to_th2(name, binedges, areas):
+            th2 = ROOT.TH2D(name, name, len(binedges[0])-1, array('d', [b for b in binedges[0]]), len(binedges[1])-1, array('d', [b for b in binedges[1]]))
+            rnp.array2hist(areas, th2, errors=errors)
+            return th2
+        th2s = {}
+        # Add nominal
+        key = key_form.format(name=self.name, variation=nominal_label)
+        th2s[key] = to_th2(key, self.binedges, self.areas)
+        for variation, areas in {**self.uncorr_variations, **self.corr_variations}.items():
+            key = key_form.format(name=self.name, variation=variation)
+            th2s[key] = to_th2(key, self.binedges, areas)
+        return th2s
+
+
+
+    def to_root_file(self, filename, recreate=False, **kwargs):
+        """Writes this histogram to a ROOT file as a set of TH2D's.
+
+        :param filename: name or path to the ROOT file. May exist or not.
+        :type filename: ``str``
+        :param recreate: only has an effect if the ROOT file exists already. If
+            `True`, it will be overwritten. Otherwise, the new histograms will
+            be added to the existing file.
+        :type recreate: ``bool``
+        :param **kwargs: keyword arguments that get passed on to
+            :py:func:`histogram2d.to_root`
+        """
+        import ROOT
+        mode = 'RECREATE' if recreate else 'UPDATE'
+        rootfile = ROOT.TFile(filename, mode)
+        if not rootfile.IsOpen():
+            raise OSError(f'Cannot open ROOT file "{filepath}"')
+        th2s = self.to_root(**kwargs)
+        for key, th2 in th2s.items():
+            th2.Write(key)
+
+
+
 class histostack(object):
     '''
     Stack of one-dimensional histograms
