@@ -5,6 +5,121 @@ import matplotlib.pyplot as plt
 from itertools import groupby
 import os
 from textwrap import dedent
+from .histogram import histogram1d as _histogram1d
+from .histogram import histogram2d as _histogram2d
+from .heatmap import _make_text, _map_to_equispaced
+from matplotlib.colors import Normalize
+
+
+
+def _make_1d_controlplot(histogram, nominal, reference, array, keys, var_hi, var_lo, path):
+    """Helper function that saves a controlplot for a 1-dimensional histogram."""
+    fig, ax = plt.subplots(1)
+    x = histogram.points()[0]
+    linestyles = ['-', '--', '-.', ':']
+    if len(x) == 1:
+        x = [0., 1.]
+        for row_index in range(array.shape[0]):
+            # There are 10 different colours in the default sequence, so change line style every 10 uncertainties to keep them distinguishable
+            linestyle = linestyles[(row_index % 40) // 10] # index switches 0, 1, 2, 3, 0, 1, 2, 3, 0 etc. every 10 steps
+            ax.plot(x, [array[row_index][0]/nominal[0], array[row_index][0]/nominal[0]], linestyle=linestyle, alpha=0.5, label=keys[row_index])
+        ax.plot(x, [reference[0]/nominal[0], reference[0]/nominal[0]], 'r-.', label='Reference')
+        ax.plot(x, [nominal[0]/nominal[0], nominal[0]/nominal[0]], 'k:', label='Nominal')
+        ax.fill_between(x, [var_hi[0]/nominal[0], var_hi[0]/nominal[0]], [var_lo[0]/nominal[0], var_lo[0]/nominal[0]], facecolor='0.75', edgecolor='0.75', alpha=0.3, linewidth=0, label='Combined (incl. postprocessing)')
+        ax.xaxis.set_ticklabels([])
+        ax.set_xlabel('Integrated', ha='right', x=1., size='x-large')
+    else:
+        for row_index in range(array.shape[0]):
+            # There are 10 different colours in the default sequence, so change line style every 10 uncertainties to keep them distinguishable
+            linestyle = linestyles[(row_index % 40) // 10] # index switches 0, 1, 2, 3, 0, 1, 2, 3, 0 etc. every 10 steps
+            ax.plot(x, array[row_index]/nominal, linestyle=linestyle, alpha=0.5, label=keys[row_index])
+        ax.plot(x, reference/nominal, 'r-.', label='Reference')
+        ax.plot(x, nominal/nominal, 'k:', label='Nominal')
+        ax.fill_between(x, var_hi/nominal, var_lo/nominal, facecolor='0.75', edgecolor='0.75', alpha=0.3, linewidth=0, label='Combined (incl. postprocessing)')
+        xlabel = histogram.attributes.get('xlabel', 'unknown')
+        ax.set_xlabel(xlabel, ha='right', x=1., size='x-large')
+
+    fig.subplots_adjust(right=0.5)
+    ax.legend(bbox_to_anchor=(1.01, 1.15), fontsize='xx-small')
+    ax.set_ylabel('Ratio to nominal', ha='right', y=1., size='x-large')
+    # print(self.controlplot + ' {0}'.format(array.shape[0]))
+    fig.savefig(path)
+    plt.close(fig)
+
+
+
+def _make_2d_controlplot(histogram, nominal, reference, array, keys, var_hi, var_lo, path):
+    """Helper function that saves a controlplot for a 2-dimensional histogram.
+
+    Caution: since there's one panel (= subplot) per variation in this plot, it will
+    become very large if there are lots of variations.
+    """
+    nrows = int(np.ceil(len(keys)/2))
+    fig, axes = plt.subplots(nrows, 2, figsize=(10, 4*nrows), squeeze=False, constrained_layout=True)
+    # fig, axes = plt.subplots(len(keys), figsize=(8, 5*len(keys)))
+    for index, (key, varied_areas) in enumerate(zip(keys, array)):
+
+        these_axes = axes[np.unravel_index(index, axes.shape)]
+
+        #print(key, varied_areas)
+        uncertainty_hist = deepcopy(histogram)
+        uncertainty_hist.corr_variations = {}
+        uncertainty_hist.uncorr_variations = {}
+        uncertainty_hist.areas = (varied_areas - nominal) / nominal * 100.0 # for percentage!
+        #print(uncertainty_hist.areas)
+
+        midx, midy, heights = uncertainty_hist.points()
+
+        # Visualised bin edges and midpoints
+        vis_binedges = uncertainty_hist.binedges
+        vis_midx = midx
+        vis_midy = midy
+        # We use monowidth visualisation, where the actual bin edges and midpoints are mapped to equidistant
+        # visualised ones:
+        # Bin edges at integers from zero:
+        vis_binedges = (np.array([i for i, _ in enumerate(uncertainty_hist.binedges[0])]), np.array([i for i, _ in enumerate(uncertainty_hist.binedges[1])]))
+        # Bin midpoints at integers from zero PLUS 0.5 (to make them midpoints w.r.t. the edges):
+        vis_midx = _map_to_equispaced(midx)
+        vis_midy = _map_to_equispaced(midy)
+
+        # midx and midy are used here as "dummy" fill values at the bin centres.
+        # The histogram is filled "with" these values with the actual bin height (or optionally area) as "weight"
+        nominals = np.ravel(uncertainty_hist.areas)
+
+        these_axes.hist2d(vis_midx, vis_midy, weights=nominals, bins=vis_binedges, cmap='bwr', norm=Normalize(vmin=-2.0, vmax=2.0))
+
+        # Write bin contents as text onto the plot
+        def text_formatter(nominal=None, **ignore):
+            text = f'{nominal:+.2f}'
+            if 'nan' in text.lower():
+                return ''
+            return text
+        for x, y, uncertainty_value in zip(vis_midx, vis_midy, nominals):
+            text = _make_text(text_formatter, {'nominal' : uncertainty_value})
+            these_axes.text(x, y, text, verticalalignment='center', horizontalalignment='center', color='black', fontsize=8)
+
+        # Set proper tick labels for monowidth plots:
+        these_axes.set_xticks(vis_binedges[0])
+        these_axes.set_yticks(vis_binedges[1])
+        # Nicely format the actual binedges: drop any trailing decimal points and zeros
+        # if all have integer values by converting the values to ints:
+        actual_binedges_x = uncertainty_hist.binedges[0].astype(int) if np.all(np.equal(np.mod(uncertainty_hist.binedges[0], 1), 0)) else np.around(uncertainty_hist.binedges[0], 5)
+        actual_binedges_y = uncertainty_hist.binedges[1].astype(int) if np.all(np.equal(np.mod(uncertainty_hist.binedges[1], 1), 0)) else np.around(uncertainty_hist.binedges[1], 5)
+        these_axes.set_xticklabels(actual_binedges_x)
+        these_axes.set_yticklabels(actual_binedges_y)
+
+        # Make the plot prettier
+        these_axes.set_title(key, x=0., ha='left', size='large')
+        if 'xlabel' in uncertainty_hist.plot_attributes:
+            these_axes.set_xlabel(uncertainty_hist.plot_attributes['xlabel'], ha='right', x=1., size='large')
+        if 'ylabel' in uncertainty_hist.plot_attributes:
+            these_axes.set_ylabel(uncertainty_hist.plot_attributes['ylabel'], ha='right', y=1., size='large')
+        these_axes.tick_params(which='both', bottom=True, top=True, left=True, right=True, direction='in')
+
+    fig.suptitle('Relative uncertainties (%)', ha='center', size='x-large')
+    # fig.subplots_adjust(left=0.08, bottom=0.08, right=0.08, top=0.08, wspace=0.08, hspace=0.08)
+    fig.savefig(path)
+
 
 
 # Terminology used in the combination methods:
@@ -44,6 +159,9 @@ def combine_asym_rms(array, nominal):
     return (hi, lo)
 
 def combine_envelope(array, nominal):
+    # This reshaping seems to be done implicitly for 1D arrays, but must be done
+    # explicitly for 2D arrays:
+    nominal = deepcopy(nominal).reshape(1, *nominal.shape)
     array_incl_nominal = np.vstack((array, nominal)) # append nominal to array
     hi = np.max(array_incl_nominal, axis=0)
     lo = np.min(array_incl_nominal, axis=0)
@@ -151,6 +269,7 @@ class model(object):
         self.matches_required = matches_required
         self.combination_functions = {
             'no_comb' : None,
+            'rename' : None,
             'drop' : None,
             'add_quad' : combine_add_quad,
             'add_lin' : combine_add_lin,
@@ -195,7 +314,7 @@ class model(object):
         keys = fnmatch.filter(list(all_keys), self.keys)
         if self.matches_required and len(keys) != self.matches_required:
             raise RuntimeError('Uncertainty combination model "{0}" cannot be applied: require {1} keys in histogram.corr_variations matching expression "{2}", but found {3}.'.format(self.name, self.matches_required, self.keys, len(keys)))
-        return keys
+        return sorted(keys)
 
 
 
@@ -221,25 +340,9 @@ class model(object):
         reference = nominal if not self.reference_key else histogram.corr_variations[self.reference_key]
         # Make array in which rows correspond to the variations:
         array = _make_variation_array(histogram.corr_variations, keys)
-        # Make controlplot if requested
-        if controlplot_location:
-            fig, ax = plt.subplots(1)
-            x = histogram.points()[0]
-            linestyles = ['-', '--', '-.', ':']
-            for row_index in range(array.shape[0]):
-                # There are 10 different colours in the default sequence, so change line style every 10 uncertainties to keep them distinguishable
-                linestyle = linestyles[(row_index % 40) // 10] # index switches 0, 1, 2, 3, 0, 1, 2, 3, 0 etc. every 10 steps
-                ax.plot(x, array[row_index]/nominal, linestyle=linestyle, alpha=0.5, label=keys[row_index])
-            ax.plot(x, reference/nominal, 'r-.', label='Reference')
-            ax.plot(x, nominal/nominal, 'k:', label='Nominal')
-            fig.subplots_adjust(right=0.5)
-            ax.legend(bbox_to_anchor=(1.01, 1.15), fontsize='xx-small')
-            # print(self.controlplot + ' {0}'.format(array.shape[0]))
-            fig.savefig(os.path.join(controlplot_location, self.controlplot))
-            plt.close(fig)
         # Combine original variations into new variations:
         combination_function = self.combination_functions[self.strategy]
-        (var_hi, var_lo) = combination_function(array, reference) # NOTE: the histogram @reference is used as the @nominal in the combination function!!!
+        var_hi, var_lo = combination_function(array, reference) # NOTE: the histogram @reference is used as the @nominal in the combination function!!!
         # Apply postprocessing if desired:
         if self.postprocess:
             if isinstance(self.postprocess, float):
@@ -253,6 +356,13 @@ class model(object):
         # Add new variations to the histogram:
         histogram.corr_variations[self.name + self.suffixes[0]] = var_hi
         histogram.corr_variations[self.name + self.suffixes[1]] = var_lo
+        # Make controlplot if requested
+        if controlplot_location:
+            controlplot_args = [histogram, nominal, reference, array, keys, var_hi, var_lo, os.path.join(controlplot_location, self.controlplot)]
+            if isinstance(histogram, _histogram1d):
+                _make_1d_controlplot(*controlplot_args)
+            elif isinstance(histogram, _histogram2d):
+                _make_2d_controlplot(*controlplot_args)
         # Drop the original variations from the input histogram:
         for key in keys:
             histogram.corr_variations.pop(key)
